import java.util.ArrayList;

/**
 *  This class is the main class of the "World of Zuul" application. 
 *  "World of Zuul" is a very simple, text based adventure game.  Users 
 *  can walk around some scenery. That's all. It should really be extended 
 *  to make it more interesting!
 * 
 *  To play this game, create an instance of this class and call the "play"
 *  method.
 * 
 *  This main class creates and initialises all the others: it creates all
 *  rooms, creates the parser and starts the game.  It also evaluates and
 *  executes the commands that the parser returns.
 * 
 * @author  Michael Kolling and David J. Barnes
 * @version 1.0 (February 2002)
 */

public class Game 
{
    private Parser parser;
    private Room currentRoom;
    private ArrayList inventory;

//  public Room valveroom;
    
    private static Game theGame = null;
        
      
//    boolean serpent_happy = false;


    public ArrayList getInventory() {
        return inventory;
    }
    
    /**
     * Create the game and initialise its internal map.
     */
    public Game() 
    {
        inventory = new ArrayList();
        createRooms();
        parser = new Parser();
        theGame = this;
    }

    /**
     * Create all the rooms and link their exits together.
     */
    private void createRooms()
    {        
    Room  outside, pub, lab, office,   
        preproom = null,
        theatre,
        steamtunnel;
    /*
     *                           zoolab              preproom
     *                              |                   |
     *                              |                   |
     *       pub-----------------outside -------------theatre
     *                              |
     *                              |
     *                              |
     *                              |
     *                             lab--------------office
     *                             
     *              steamtunnel -------- basement
     *                             
     */
        // create the rooms
        outside = new Room("outside the main entrance of the university", this);
        outside.addItem(new Item("key"));       // cheat
        
        theatre = new Room("in a lecture theatre");
        theatre = new LockedRoom("in a lecture theatre", "north", inventory);
        theatre.addItem(new Item("pencil"));
        
        pub = new Room("in the campus pub");
        pub.addItem(new Item("water", 3));
        
        lab = new Room("in a computing lab");
        
        office = new Room("in the computing admin office");
        office.addItem(new Item("memo", 1));
        office.addItem(new Item("laptop", 7));
  //    office.addItem(new Item("cookie", 1));
        
        steamtunnel = new SerpentRoom("in a dark steam tunnel beneath campus"
                                + "\nThe door just snapped shut above you"
                                + "\nA giant serpent is coiled before you!!!");
        
        Room basement1 = new Room("in a dark basement somewhere");
        Room basement2 = new Room("in a basement somewhere. Dim light and a strange hum"
                            + "\ncomes from the north");

        Room basementlab = null;
        Room valveroom = null;
        Room bottomroom = null;
        Room rocktunnel = null;
        Room basement3 = null;
        
        link(steamtunnel, "east", basement1, "west");
//        steamtunnel.setExit("east", basement1);
//        basement1.setExit("west", steamtunnel);

        link(basement1, "north", basement2, "south");
//        basement1.setExit("north", basement2);
//        basement2.setExit("south", basement1);
        
        
        basementlab = new Room("in a huge underground chamber that contains enormous apparatus"
                                + "\nfrom a long-ago experiment. At one end of the room"
                                + "\nis a pool of water that, in the dim light, looks very deep."
                                +"\nThere is also a damp metal stairway leading down");
//        basementlab.setExit("east", basement3);
        basementlab.addItem(new Item("glowing_donut"));

        link(basement2, "north", basementlab, "south");

        valveroom = new ValveRoom(true, "down",
                                "in a deeper part of the basement, below the experiment chamber."
                            +   "\nThere are many pipes here, and a huge valve.",
                            "You find that the steps are flooded with water"
                                +"\nYou need to find a way to drain the water"
                            );
        
        link(basementlab, "down", valveroom, "up");                    
                  
        bottomroom = new Room("you are deep in the sub-basement"
                             +"\nThe bare-rock walls are dripping with water."
                             +"\nA tunnel angles south and slightly upwards");
                             
        link(valveroom, "down", bottomroom, "up");                    
//      rocktunnel = new Room("in a long sloping rock tunnel, very wet at the north end"
//                            + "\nA ladder leads up from the middle");
        rocktunnel = new ValveRoom(false, "north",
                                "in a long sloping rock tunnel, very wet at the north end",
                            "Hmm...the north end is MUCH too wet to allow progress"
                            );
        
        Room darkroom = new DarkRoom("north", 
                                     "A cavern room with stalactites and stalagmites",
                                     inventory);
        darkroom.addItem(new Item("goldcoin"));              // hmm... will this work?
        Room gemroom = new Room("A small cavern room");
        gemroom.addItem(new Item("bag_of_gems"));
                
        link(bottomroom, "south", rocktunnel, "north");
        link(rocktunnel, "south", darkroom, "north");
        link(darkroom, "south", gemroom, "north");
        
        basement3 = new Room("in a crawlspace under the lecture room"
                        + "\nAbove you are the undersides of the seats");
        link(rocktunnel, "up", basement3, "down");
                        
        
        
//=====================================================================

        preproom = new Room("Lecture preparation room");
        preproom.addItem(new Item("tuition_vouchers"));
 //       basement3.setExit("up", preproom);
        link(basement3, "up", preproom, "down");
        
        Room zoolab = new ZooRoom("You are in the zoological laboratory" 
                                + "\nThere are cages everywhere."
                                );
        // initialise room exits
        outside.setExit("east", theatre);
        outside.setExit("south", lab);
        outside.setExit("west", pub);
        link(outside, "north", zoolab, "south");

        theatre.setExit("west", outside);
        theatre.setExit("north", preproom);

        pub.setExit("east", outside);

        lab.setExit("north", outside);
        lab.setExit("east", office);
        lab.setExit("down", steamtunnel);

        office.setExit("west", lab);
        
        steamtunnel.setExit("up", pub);
        // steamtunnel.addItem(new Item("key"));    // added dynamically LATER!!!
        
        preproom.setExit("south", theatre);

        currentRoom = outside;  // start game outside
        
        inventory.add(new Light("flashlight", 4));
        inventory.add(new Item("cookie", 2));
//      inventory.add(new Item("cellphone"));
    }

    /**
     *  Main play routine.  Loops until end of play.
     */
    public void play() 
    {            
        printWelcome();

        // Enter the main command loop.  Here we repeatedly read commands and
        // execute them until the game is over.
                
        boolean finished = false;
        while (! finished) {
            Command command = parser.getCommand();
            finished = processCommand(command);
        }
        System.out.println("Thank you for playing.  Good bye.");
    }

    /**
     * Print out the opening message for the player.
     */
    private void printWelcome()
    {
        System.out.println();
        System.out.println("Welcome to the World of Zuul!");
        System.out.println("World of Zuul is a new, incredibly boring adventure game.");
        System.out.println("Type 'help' if you need help.");
        System.out.println();
        currentRoom.onEntry();
        System.out.println(currentRoom.getLongDescription());
    }

    /**
     * Given a command, process (that is: execute) the command.
     * If this command ends the game, true is returned, otherwise false is
     * returned.
     */
    private boolean processCommand(Command command) 
    {
        boolean wantToQuit = false;

        if(command.isUnknown()) {
            System.out.println("I don't know what you mean...");
            return false;
        }

        String commandWord = command.getCommandWord();
        
        if (commandWord.equals("quit")) {       // serpent
            return true;        
        }
        
       
        if (commandWord.equals("help"))
            printHelp();
        else if (commandWord.equals("go")) {
            {
                goRoom(command);
            }
        } else if (commandWord.equals("inventory")) {
            printInventory();
        } else if (commandWord.equals("look")) {
            System.out.println(currentRoom.getLongDescription());
        } else {
            currentRoom.respond(command, inventory);
        }
              
        return wantToQuit;      // always false
    }


    /**
     * Print out some help information.
     * Here we print some stupid, cryptic message and a list of the 
     * command words.
     */
    private void printHelp() 
    {
        System.out.println("You are lost. You are alone. You wander");
        System.out.println("around at the university.");
        System.out.println();
        System.out.println("Your command words are:");
        parser.showCommands();
    }

    /** 
     * Try to go to one direction. If there is an exit, enter the new
     * room, otherwise print an error message.
     * pld: if getExit() returns null, we can't go that way, and getExit()
     * will print why. Returning a null signals to goRoom *not* to
     * print the description. 
     * Note that there may be cases where getExit returns the same room as the one we're in;
     * the usual understanding of such a situation is that we somehow just returned
     * to the room we were in (as via a teleporter or in a maze), but it wasn't
     * "obviously" the same room.
     */
    private void goRoom(Command command) 
    {
        if(!command.hasSecondWord()) {
            // if there is no second word, we don't know where to go...
            System.out.println("Go where?");
            return;
        }

        String direction = command.getSecondWord();

        // Try to leave current room.
        Room nextRoom = currentRoom.getExit(direction);

        if (nextRoom == null) {
            // Make getExit be responsible for printing out the reason why not - pld
            //System.out.println("There is no door!");
        } else {
            currentRoom = nextRoom;
            currentRoom.onEntry();
            System.out.println(currentRoom.getLongDescription());
        }
    }

    /** 
     * "Quit" was entered. Check the rest of the command to see
     * whether we really quit the game. Return true, if this command
     * quits the game, false otherwise.
     */
    private boolean quit(Command command) 
    {
        if(command.hasSecondWord()) {
            System.out.println("Quit what?");
            return false;
        }
        else
            return true;  // signal that we want to quit
    }
    
    /**
     * printInventory prints the contents of "inventory"
     */
    private void printInventory() {
        if (inventory.size() == 0) {
            System.out.println("you are not carrying anything");
        } else {
            System.out.print("You have the following:");
            for (int n = 0; n < inventory.size(); n++) {
                Item item = (Item) inventory.get(n);
                System.out.print(" " + item.getDesc());
            }
            System.out.println();
            System.out.println("Total weight: " + totalWeight(inventory));
        }
    }
    
    /**
     * findByName: given a string and an ArrayList of Items, 
     * find the Item with the matching name, or else return null
     */
    static public Item findByName(String s, ArrayList L) {
        int n=0;
        while (n < L.size()) {
            Item i = (Item) L.get(n);
            if (s.equals(i.getDesc())) 
                return i;
            n++;
        }
        return null;    // not found above
    }
   
   /**
     * totalWeight: given an ArrayList of Items, returns total of weights
     */
    public int totalWeight(ArrayList L) {
        int n=0;
        int sum = 0;
        while (n < L.size()) {
            Item i = (Item) L.get(n);
            sum += i.getWeight();
            n++;
        }
        return sum;    // not found above
    }
    
    /* */
    public static void main(String [] args) {
        initGame();
        Game g = getGame();
        g.play();
    }
    
    /**
     * initGame(): like ctor, except checks that there isn't already one
     */
    public static void initGame() {
        if (theGame != null) {
            theGame = new Game();
        }
    }
    
    /** 
     * for access to the game from outside.
     * Problem: what if there are multiple games?
     */
    public static Game getGame() {
        return theGame;
    }
    
    /**
     * link: creates a link between rooms r1 and r2:
     * From r1, going dir1, we get to r2, and from r2, going dir2, we get to r1
     */
    private void link(Room r1, String dir1, Room r2, String dir2) {
        r1.setExit(dir1, r2);
        r2.setExit(dir2, r1);
    }
     
    /* */
}
